unit hiListBoxEx;

interface

uses
  Kol,Windows,Messages,Share,Debug,Win;

const
  WM_LISTBOX_INVALIDATE = WM_USER + $2000;

type
  TListBoxDrawItem = procedure(Sender: PControl; aCanvas: PCanvas;
                        const aRect: TRect; aIndex: Integer) of object;

  TInplaceEdit = class;

  THIListBoxEx = class(THIWin)
  private
    FEdit: TInplaceEdit;
    FList: PStrList;
    FHideSelection: Boolean;
    FMinItemHeight: Integer;
    FCount: Integer;
    FItemIndex: Integer;
    FTopIndex: Integer;
    FVisibleCount: Integer; //    
    FItemHeight: Integer;

    OldMes: TOnMessage;
    FFocusRect: TRect;
    FOnDrawItem: TListBoxDrawItem;
    Arr: PArray;

    procedure _Set(var Item:TData; var Val:TData);
    function _Get(Var Item:TData; var Val:TData):boolean;
    function _Count:integer;
    procedure _Add(var Val:TData);

    function  CorrectItemIndex(Value: Integer): Integer;
    function  CorrectTopIndex(Value: Integer): Integer;
    procedure SetCount(Value: Integer);
    procedure SetItemIndex(Value: Integer);
    procedure SetTopIndex(Value: Integer);
    procedure SetMinItemHeight(Value: Integer);
    procedure SetItemHeight;
    procedure AdjustParams;

    procedure Resize(Obj: PObj);
    procedure Paint(Sender: PControl; DC: HDC);
    procedure ListBoxDrawItem(Sender: PControl; aCanvas: PCanvas;
                        const aRect: TRect; aIndex: Integer);

    procedure MouseDown(Sender: PControl; var Mouse: TMouseEventData);
    procedure KeyDown(Sender: PControl; var Key: Longint; Shift: DWORD);
    procedure KeyUp(Sender: PControl; var Key: Longint; Shift: DWORD);
    procedure MouseWheel(Sender: PControl; var Mouse: TMouseEventData);
    function OnMessage( var Msg: TMsg; var Rslt: Integer ): Boolean;


    property HideSelection: Boolean read FHideSelection write FHideSelection default True;
    property MinItemHeight: Integer read FMinItemHeight
             write SetMinItemHeight default 0;
    property OnDrawItem: TListBoxDrawItem read FOnDrawItem write FOnDrawItem;

  public
    _prop_FileName: string;
    _prop_Strings: string;
    _prop_AddType: byte;
    _prop_DataType: byte;
    _prop_EditMode: boolean;

    _data_Str: THI_Event;
    _data_FileName: THI_Event;

    _event_onClick: THI_Event;
    _event_onChange: THI_Event;

    _event_onStartEdit: THI_Event;
    _event_onEndEdit: THI_Event;

    procedure Init;
    procedure SetItemIndexCentered(Value: Integer);
    function  HitTest(Y: Integer): Integer;

    procedure _work_doAdd(var _Data: TData; Index: word);
    procedure _work_doClear(var _Data: TData; Index: word);
    procedure _work_doDelete(var _Data: TData; Index: word);
    procedure _work_doText(var _Data: TData; Index: word);
    procedure _work_doSelect(var _Data: TData; Index: word);
    procedure _work_doSelectString(var _Data: TData; Index: word);
    procedure _work_doReplace(var _Data: TData; Index: word);
    procedure _work_doLoad(var _Data: TData; Index: word);
    procedure _work_doSave(var _Data: TData; Index: word);


    procedure _var_Text(var _Data: TData; Index: word);
    procedure _var_Count(var _Data: TData; Index: word);
    procedure _var_String(var _Data: TData; Index: word);
    procedure _var_Index(var _Data: TData; Index: word);
    procedure _var_Array(var _Data:TData; Index: word);

    property Count: Integer read FCount write SetCount;
    property ItemIndex: Integer read FItemIndex write SetItemIndex;
    property TopIndex: Integer read FTopIndex write SetTopIndex;
    property VisibleCount: Integer read FVisibleCount;

  end;

  TInplaceEdit = class(TObject)
  private
    FEdit: PControl;
    FStyle: Integer;
    FListBox: THIListBoxEx;
    OldMes: TOnMessage;
  protected
    procedure Change(Obj:PObj);
    procedure KeyUp(Sender: PControl; var Key: Longint; Shift: DWORD);
    function OnMessage( var Msg: TMsg; var Rslt: Integer ): Boolean;
  public
    constructor Create(ListBox: THIListBoxEx);
    procedure ShowEdit(Selection: TRect);
    procedure HideEdit;
  end;

implementation

procedure THIListBoxEx._work_doReplace(var _Data: TData; Index: word);
var
  s: string;
  idx: integer;
begin
  idx := ToInteger(_Data);
  s := ToStringEvent(_data_Str);
  if (idx >= 0) and (idx < FList.Count) then
  begin
    FList.Items[idx] := s;
    Control.Invalidate;
    _hi_OnEvent(_event_onChange);
  end;
end;

procedure THIListBoxEx._work_doSelectString(var _Data: TData; Index: word);
var
  s: string;
begin
  s := ToString(_Data);
  if s <> '' then
    SetItemIndexCentered(FList.IndexOf(s));
end;

procedure THIListBoxEx._work_doSelect(var _Data: TData; Index: word);
var
  idx: integer;
begin
  idx := ToInteger(_Data);
  if (idx >= 0) and (idx < FList.Count) then
    SetItemIndexCentered(idx);
end;

procedure THIListBoxEx._work_doAdd(var _Data: TData; Index: word);
var
  s: string;
begin
  s := ReadString(_Data,_data_Str,'');
  if s <> '' then
  begin
    if _prop_AddType = 0 then
      FList.Add(s);
    if _prop_AddType = 1 then
      FList.Insert(0,s);

    Inc(FCount);
    AdjustParams;
    Control.Invalidate;
    _hi_OnEvent(_event_onChange);
  end;
end;

procedure THIListBoxEx._work_doClear(var _Data: TData; Index: word);
begin
  FList.Clear;
  FCount := 0;
  AdjustParams;
  Control.Invalidate;
  _hi_OnEvent(_event_onChange);
end;

procedure THIListBoxEx._work_doDelete(var _Data: TData; Index: word);
var
  idx: integer;
begin
  idx := ToInteger(_Data);
  if (idx >= 0) and (idx < FList.Count) then
  begin
    FList.Delete(idx);
    Dec(FCount);
    AdjustParams;
    Control.Invalidate;
    _hi_OnEvent(_event_onChange);
  end;
end;

procedure THIListBoxEx._work_doText(var _Data: TData; Index: word);
begin
  FList.Text := ToString(_Data);
  FCount := FList.Count-1;
  AdjustParams;
  Control.Invalidate;
  _hi_OnEvent(_event_onChange);
end;

procedure THIListBoxEx._work_doLoad(var _Data: TData; Index: word);
var
  FileName: string;
begin
  FileName := ReadString(_Data,_data_FileName,_prop_FileName);
  FList.LoadFromFile(FileName);
end;

procedure THIListBoxEx._work_doSave(var _Data: TData; Index: word);
var
  FileName: string;
begin
  FileName := ReadString(_Data,_data_FileName,_prop_FileName);
  FList.SaveToFile(FileName);
end;

procedure THIListBoxEx._var_Text(var _Data: TData; Index: word);
begin
  dtString(_Data,FList.Text)
end;

procedure THIListBoxEx._var_Count(var _Data: TData; Index: word);
begin
  dtInteger(_Data,FList.Count);
end;

procedure THIListBoxEx._var_String(var _Data: TData; Index: word);
begin
  dtString(_Data,FList.Items[ItemIndex]);
end;

procedure THIListBoxEx._var_Index(var _Data: TData; Index: word);
begin
  dtInteger(_Data,ItemIndex);
end;


{ THIListBoxEx }
procedure THIListBoxEx.ListBoxDrawItem(Sender: PControl; aCanvas: PCanvas;
                        const aRect: TRect; aIndex: Integer);
var
  R: TRect;
begin
  if aIndex = FItemIndex then
  begin
    FFocusRect := aRect;
    if Control.Focused then
      aCanvas.DrawFocusRect(aRect);
  end;
  R := aRect;
  aCanvas.DrawText(FList.Items[aIndex],R, DT_LEFT);
end;

function THIListBoxEx.CorrectItemIndex(Value: Integer): Integer;
begin
  if Value < 0 then
    Result := 0
  else if Value > (FCount - 1) then
    Result := FCount - 1
  else
    Result := Value;
end;

function THIListBoxEx.CorrectTopIndex(Value: Integer): Integer;
begin
  if Value < 0 then
    Result := 0
  else if (Value + FVisibleCount) > FCount then begin
    Result := FCount - FVisibleCount;
    if Result < 0 then
      Result := 0;
  end
  else
    Result := Value;
end;

procedure THIListBoxEx.SetCount(Value: Integer);
begin
  if Value < 0 then
    Value := 0;
  if FCount <> Value then begin
    FCount := Value;
    AdjustParams;
  end;
end;

procedure THIListBoxEx.SetItemIndex(Value: Integer);
var
  ti:  Integer;
  inv: Boolean;
begin
  if FItemHeight = 0 then
    SetItemHeight;
  ti := FTopIndex;
  Value := CorrectItemIndex(Value);
  if Value < ti then
    ti := Value
  else if Value >= (ti + FVisibleCount) then
    ti := Value - FVisibleCount + 1;
  ti := CorrectTopIndex(ti);
  inv := False;
  if FItemIndex <> Value then
  begin
    FItemIndex := Value;
    inv := True;
  end;
  if FTopIndex <> ti then begin
    FTopIndex  := ti;
    SetScrollPos(Control.Handle, SB_VERT, FTopIndex, True);
    inv := True;
  end;
  if inv then
    Control.Invalidate;
  if _prop_DataType = 0 then
    _hi_OnEvent(_event_OnClick,ItemIndex);
  if _prop_DataType = 1 then
    _hi_OnEvent(_event_OnClick,FList.Items[ItemIndex]);
end;

procedure THIListBoxEx.SetTopIndex(Value: Integer);
begin
  if FItemHeight = 0 then
    SetItemHeight;
  Value := CorrectTopIndex(Value);
  if FTopIndex <> Value then begin
    FTopIndex := Value;
    SetScrollPos(Control.Handle, SB_VERT, FTopIndex, True);
    Control.Invalidate;
  end;
end;

procedure THIListBoxEx.SetMinItemHeight(Value: Integer);
begin
  if FMinItemHeight <> Value then begin
    FMinItemHeight := Value;
    FItemHeight    := 0;
    AdjustParams;
  end;
end;

procedure THIListBoxEx.SetItemHeight;
begin
  FItemHeight := Control.Canvas.TextHeight('M');
  if FItemHeight = 0 then
    FItemHeight := 16;
  if FItemHeight < FMinItemHeight then
    FItemHeight := FMinItemHeight;
  FVisibleCount := Control.ClientHeight div FItemHeight;
end;

procedure THIListBoxEx.AdjustParams;
var
  si: TScrollInfo;
begin
  if FItemHeight = 0 then
    SetItemHeight;
  if FCount = 0 then begin
    FItemIndex := -1;
    FTopIndex  := 0;
  end
  else begin
    FItemIndex := CorrectItemIndex(FItemIndex);
    FTopIndex  := CorrectTopIndex(FTopIndex);
  end;
  if FCount > FVisibleCount then begin
    ZeroMemory(@si, SizeOf(si));
    si.cbSize := SizeOf(TScrollInfo);
    si.fMask  := SIF_RANGE or SIF_POS;
    si.nMin   := 0;
    si.nMax   := FCount - FVisibleCount;
    si.nPos   := FTopIndex;
    SetScrollInfo(Control.Handle, SB_VERT, si, True);
  end
  else
    ShowScrollBar(Control.Handle, SB_VERT, False);
  Control.Invalidate;
end;

function THIListBoxEx.OnMessage( var Msg: TMsg; var Rslt: Integer ): Boolean;
var
  si: TScrollInfo;
begin
  case Msg.message of
    WM_VSCROLL:
      begin
        FEdit.HideEdit;
        if LoWord(Msg.wParam) = SB_LINEUP then
          SetTopIndex(FTopIndex - 1);
        if LoWord(Msg.wParam) = SB_LINEDOWN then
          SetTopIndex(FTopIndex + 1);
        if LoWord(Msg.wParam) = SB_PAGEUP then
          SetTopIndex(FTopIndex - FVisibleCount);
        if LoWord(Msg.wParam) = SB_PAGEDOWN then
         SetTopIndex(FTopIndex + FVisibleCount);

        if LoWord(Msg.wParam) = SB_THUMBTRACK then
        begin
          ZeroMemory(@si, SizeOf(si));
          si.cbSize := SizeOf(si);
          SI.fMask  := SIF_TRACKPOS;
          if GetScrollInfo(Control.Handle, SB_VERT, si) then
            SetTopIndex(si.nTrackPos);
        end;

      end;
    WM_SETFOCUS:
      begin
        if FHideSelection then
          PostMessage(Control.Handle, WM_LISTBOX_INVALIDATE, 0, 0);
      end;
    WM_KILLFOCUS:
      begin
        if FHideSelection then
          PostMessage(Control.Handle, WM_LISTBOX_INVALIDATE, 0, 0);
      end;

    WM_LISTBOX_INVALIDATE: Control.Invalidate;

    WM_LBUTTONDBLCLK:
      begin
         //_debug('onDblClick');
         FEdit.ShowEdit(FFocusRect);
      end;
  end;
  Result := _hi_OnMessage(OldMes,Msg,Rslt);
end;

procedure THIListBoxEx.Resize;
begin
  inherited;
  FItemHeight := 0;
  AdjustParams;
end;

procedure THIListBoxEx.Paint;
var
  R: TRect;
  i, im: Integer;
  Bmp: Kol.PBitmap;
begin
  if (FCount <> 0) and Assigned(FOnDrawItem) then begin
  Bmp := NewBitmap(Control.Width,Control.Height);

  Bmp.Canvas.Brush.Color := _prop_Color;
  Bmp.Canvas.Brush.BrushStyle := bsSolid;
  Bmp.Canvas.FillRect(Control.ClientRect);
  Bmp.Canvas.Font.Assign(Control.Font);

    if FItemHeight = 0 then
      SetItemHeight;
    R := Control.ClientRect;
    im := FTopIndex + FVisibleCount + 1;
    if im >= Pred(FCount) then
      im := Pred(FCount);
    SetBkMode(Bmp.Canvas.Handle, TRANSPARENT);
    for i := FTopIndex to im do begin
      R.Bottom := R.Top + FItemHeight;
      if i = FItemIndex then begin
        if FHideSelection and (not Control.Focused) then begin
          Bmp.Canvas.Brush.Color := clBtnFace;
          Bmp.Canvas.Font.Color  := clBtnText;
        end
        else begin
          Bmp.Canvas.Brush.Color := clHighlight;
          Bmp.Canvas.Font.Color  := clHighlightText;
        end;
        Bmp.Canvas.FillRect(R);
        SetBkMode(Bmp.Canvas.Handle, TRANSPARENT);
      end;
      FOnDrawItem(Control, Bmp.Canvas, R, i);
      if i = FItemIndex then begin
        Bmp.Canvas.Brush.Color := Control.Color;
        Bmp.Canvas.Font.Color  := Control.Font.Color;
        SetBkMode(Bmp.Canvas.Handle, TRANSPARENT);
      end;
      R.Top := R.Bottom;
    end;
    Bmp.Draw(DC,0,0);
    Bmp.Free;
  end;
end;

procedure THIListBoxEx.KeyDown(Sender: PControl; var Key: Longint; Shift: DWORD);
begin
  case Key of
    VK_UP:    SetItemIndex(FItemIndex - 1);
    VK_DOWN:  SetItemIndex(FItemIndex + 1);
    VK_PRIOR: SetItemIndex(FItemIndex - FVisibleCount);
    VK_NEXT:  SetItemIndex(FItemIndex + FVisibleCount);
    VK_HOME:  SetItemIndex(0);
    VK_END:   SetItemIndex(FCount);
  end;
end;

procedure THIListBoxEx.KeyUp(Sender: PControl; var Key: Longint; Shift: DWORD);
begin
  if Key = VK_RETURN then
    FEdit.ShowEdit(FFocusRect);
end;

procedure THIListBoxEx.MouseDown(Sender: PControl; var Mouse: TMouseEventData);
begin
  FEdit.HideEdit;
  if Control.TabStop and (not Control.Focused) then
    Control.Focused := True;
  if FItemHeight = 0 then
    SetItemHeight;
  SetItemIndex(FTopIndex + (Mouse.Y div FItemHeight));
end;

procedure THIListBoxEx.MouseWheel(Sender: PControl; var Mouse: TMouseEventData);
var
  Ms: TMouseEventData;
  test: integer;
  dt: TData;
begin
   Ms := Mouse;
   test := integer(Ms.Shift) div $10000;
   if test = 120 then
     SetTopIndex(FTopIndex - 1);

   if test = -120 then
     SetTopIndex(FTopIndex + 1);

   _hi_OnEvent(_event_onMouseWheel, integer(Ms.Shift)div $10000);
end;

procedure THIListBoxEx.Init;
var
  Style: DWORD;
begin
  Style := WS_CHILD or WS_TABSTOP or WS_VISIBLE  or
  WS_VSCROLL or WS_BORDER or BS_PUSHLIKE;
  Control := _NewControl(FParent,'THIListBoxEx', Style, False,nil);
  Control.ClsStyle := Control.ClsStyle or CS_DBLCLKS;
  inherited;

  FHideSelection := True;
  FItemIndex     := -1;
  FList := NewStrList;
  FList.Text := _prop_Strings;


  Count := FList.Count;
  //FMinItemHeight := 40;

  FEdit := TInplaceEdit.Create(Self);

  if _prop_WinStyle = 0 then
    FEdit.FStyle := 2;

  if _prop_WinStyle = 1 then
    FEdit.FStyle := 0;

  if _prop_WinStyle = 2 then
    FEdit.FStyle := 1;

  if _prop_WinStyle = 3 then
    FEdit.FStyle := 3;

  if _prop_WinStyle = 4 then
    FEdit.FStyle := 5;

  Control.OnPaint := Paint;
  Control.OnMouseDown := MouseDown;
  Control.onMouseWheel := MouseWheel;
  Control.OnKeyDown := KeyDown;
  Control.OnKeyUp := KeyUp;
  Control.onResize := Resize;
  OldMes := Control.onMessage;
  Control.OnMessage := OnMessage;
  OnDrawItem := ListBoxDrawItem;
  Control.Invalidate;
end;

procedure THIListBoxEx.SetItemIndexCentered(Value: Integer);
var
  ti:  Integer;
  inv: Boolean;
begin
  if FItemHeight = 0 then
    SetItemHeight;
  Value := CorrectItemIndex(Value);
  inv   := False;
  if (Value >= FTopIndex) and (Value < (FTopIndex + FVisibleCount)) then
    ti := FTopIndex
  else
    ti := CorrectTopIndex(Value - (FVisibleCount div 2));
  if FItemIndex <> Value then begin
    FItemIndex := Value;
    inv := True;
    if _prop_DataType = 0 then
      _hi_OnEvent(_event_OnClick,ItemIndex);
    if _prop_DataType = 1 then
      _hi_OnEvent(_event_OnClick,FList.Items[ItemIndex]);
  end;
  if FTopIndex <> ti then begin
    FTopIndex  := ti;
    SetScrollPos(Control.Handle, SB_VERT, FTopIndex, True);
    inv := True;
  end;
  if inv then
    Control.Invalidate;
end;

function THIListBoxEx.HitTest(Y: Integer): Integer;
begin
  Result := FTopIndex + (Y div FItemHeight);
  if Result >= FCount then
    Result := -1;
end;

procedure THIListBoxEx._Set;
var   ind:integer;
begin
   ind := ToIntIndex(Item);
   if (ind >= 0) and (ind < FList.Count) then
      FList.Items[ind] := ToString(Val);
end;

procedure THIListBoxEx._Add;
begin
   FList.Add(ToString(val));
end;

function THIListBoxEx._Get;
var   ind:integer;
begin
   ind := ToIntIndex(Item);
   if (ind >= 0) and (ind < FList.Count) then begin
      Result := true;
      dtString(Val,FList.Items[ind]);
   end else
      Result := false;
end;

function THIListBoxEx._Count;
begin
   Result := FList.Count;
end;

procedure THIListBoxEx._var_Array;
begin
   if Arr = nil then Arr := CreateArray(_Set, _Get, _Count, _Add);
   dtArray(_Data,Arr);
end;


{ TInplaceEdit }
constructor TInplaceEdit.Create(ListBox: THIListBoxEx);
begin
  FEdit := NewEditbox(ListBox.Control.Parent,[eoNoHideSel,eoNoVScroll,eoNoHScroll]);
  //FEdit.ExStyle := WS_EX_LTRREADING;
  FEdit.Style := $56010104;
  FEdit.ExStyle := 0;
  FListBox := ListBox;

  FEdit.Left := -200;
  FEdit.Top := -200;
  FEdit.Visible := False;

  FEdit.OnKeyUp := KeyUp;
  FEdit.OnChange := Change;
  OldMes := FEdit.onMessage;
  FEdit.OnMessage := OnMessage;
end;

procedure TInplaceEdit.ShowEdit(Selection: TRect);
begin
  if not FListBox._prop_EditMode then Exit;
  _hi_onEvent(FListBox._event_onStartEdit,FListBox.FList.Items[FListBox.ItemIndex]);
  FEdit.Left := Selection.Left + FListBox.Control.Left+FStyle; //+2   ListBox
  FEdit.Top := Selection.Top + FListBox.Control.Top+FStyle;
  FEdit.Width := Selection.Right - Selection.Left;
  FEdit.Height := Selection.Bottom - Selection.Top;

  FEdit.Color := FListBox._prop_Color;//clWindow;
  FEdit.Font.Assign(FListBox.Control.Font);
  FEdit.Text := FListBox.FList.Items[FListBox.ItemIndex];
  FEdit.SelStart := Length(FEdit.Text);
  FEdit.Show;
  FEdit.BringToFront;
  FEdit.Focused := True;
  FEdit.Invalidate;
  ShowCaret(FEdit.Handle);
  FEdit.OnKeyUp := KeyUp;
end;

procedure TInplaceEdit.HideEdit;
begin
  if FEdit.Visible then
  begin
    _hi_onEvent(FListBox._event_onEndEdit,FEdit.Text);
    FEdit.Hide;
  end;
end;

procedure TInplaceEdit.Change(Obj: PObj);
begin
  FListBox.FList.Items[FListBox.ItemIndex] := FEdit.Text;
end;

procedure TInplaceEdit.KeyUp(Sender: PControl; var Key: Longint; Shift: DWORD);
begin
  case Key of
    VK_ESCAPE, VK_RETURN, VK_UP, VK_DOWN:
    begin
      FEdit.Left := -200;
      FEdit.Top := -200;
      FEdit.Text := '';
      HideEdit;
      FEdit.Focused := False;
      FEdit.OnKeyUp := nil;
      SendMessage(FListBox.Control.Handle,WM_SETFOCUS,0,0);
    end;
  end;
end;

function TInplaceEdit.OnMessage( var Msg: TMsg; var Rslt: Integer ): Boolean;
begin
  case Msg.message of
    WM_KILLFOCUS: HideEdit;
  end;
  Result := _hi_OnMessage(OldMes,Msg,Rslt);
end;

end.
